/* HPCU Bi-Weekly Challenge: Diffusion Limited Aggregation with NetCDF
 * Starter code
 * Aaron Weeden, Shodor, 2016
 */

// Import libraries
#include <stdbool.h>
#include <stdlib.h>

// Set model parameters
#define PARTICLE_COUNT     201  // # of particles in the simulation
#define TIME_COUNT         1000 // # of time steps in the simulation
#define ENVIRONMENT_LENGTH 100  /* Length of one side of the square environment,
                                   measured in particle lengths */
#define STICKINESS         100  // (0-100) likelihood of a particle sticking

// Define constants for particle states
#define NONE  0
#define FREE  1
#define STUCK 2

// Define a type for movement directions
typedef enum {
  UP, LEFT, RIGHT, DOWN
} Direction;

// Declare global arrays to hold the particle locations and states
int ParticleXs[PARTICLE_COUNT];
int ParticleYs[PARTICLE_COUNT];
int ParticleStates[PARTICLE_COUNT];
int Environment[ENVIRONMENT_LENGTH][ENVIRONMENT_LENGTH];

// Define a function that returns a random integer less than a given maximum
int GetRandom(int max) {
  return max * random() / RAND_MAX; 
}

// Define a function for initializing particle locations and states
void Init() {
  // Set the environment as having no state
  for (int y = 0; y < ENVIRONMENT_LENGTH; y++) {
    for (int x = 0; x < ENVIRONMENT_LENGTH; x++) {
      Environment[y][x] = NONE;
    }
  }

  // Center and stick the first particle
  ParticleXs[0] = ParticleYs[0] = ENVIRONMENT_LENGTH / 2 - 1;
  ParticleStates[0] = STUCK;
  Environment[ParticleYs[0]][ParticleXs[0]] = STUCK;

  // Loop over the other particles, give them random positions, and make them
  // free
  for (int i = 1; i < PARTICLE_COUNT; i++) {
    ParticleXs[i] = GetRandom(ENVIRONMENT_LENGTH);
    ParticleYs[i] = GetRandom(ENVIRONMENT_LENGTH);
    ParticleStates[i] = FREE;
    Environment[ParticleYs[i]][ParticleXs[i]] = FREE;
  }
}

// Define a function that sticks free particles
void StickParticles() {
  // Loop over all particles
  for (int i = 0; i < PARTICLE_COUNT; i++) {

    // If the particle is stuck, ignore it
    if (ParticleStates[i] == STUCK) {
      continue;
    }

    // Loop over all the other particles
    for (int j = 0; j < PARTICLE_COUNT; j++) {

      // If the particle is free, ignore it
      if (ParticleStates[j] == FREE) {
        continue;
      }

      // If the two particles are not adjacent, ignore them
      if (ParticleXs[i] > ParticleXs[j] + 1 ||
          ParticleXs[j] > ParticleXs[i] + 1 ||
          ParticleYs[i] > ParticleYs[j] + 1 ||
          ParticleYs[j] > ParticleYs[i] + 1) {
        continue;
      }

      // Stick the particle with some % chance
      if (GetRandom(100) < STICKINESS) {
        ParticleStates[i] = STUCK;
        Environment[ParticleYs[i]][ParticleXs[i]] = STUCK;
      }
    }
  }
}

// Define a function that returns whether a particle, moving in a given
// direction, would collide with another particle
bool wouldCollide(int const i, Direction const dir) {
  for (int j = 0; j < PARTICLE_COUNT; j++) {
    // Do not check for a particle colliding with itself
    if (i == j) {
      continue;
    }

    if ((dir == UP &&
         ParticleXs[j] == ParticleXs[i] &&
         ParticleYs[j] == ParticleYs[i] - 1) ||
        (dir == LEFT &&
         ParticleYs[j] == ParticleYs[i] &&
         ParticleXs[j] == ParticleXs[i] - 1) ||
        (dir == RIGHT &&
         ParticleYs[j] == ParticleYs[i] &&
         ParticleXs[j] == ParticleXs[i] + 1) ||
        (dir == DOWN &&
         ParticleXs[j] == ParticleXs[i] &&
         ParticleYs[j] == ParticleYs[i] + 1)) {
      return true;
    }
  }
  return false;
}

// Define a function that moves free particles randomly
void MoveRandom() {
  // Loop over all particles
  for (int i = 0; i < PARTICLE_COUNT; i++) {
    // If the particle is stuck, skip it
    if (ParticleStates[i] == STUCK) {
      continue;
    }

    // Get random direction
    Direction dir;
    switch (GetRandom(4)) {
      case 0:
        dir = UP;
        break;
      case 1:
        dir = LEFT;
        break;
      case 2:
        dir = RIGHT;
        break;
      default:
        dir = DOWN;
        break;
    }

    // If the move would cause a collision with a wall, skip this particle
    if ((dir == UP    && ParticleYs[i] <= 0) ||
        (dir == LEFT  && ParticleXs[i] <= 0) ||
        (dir == RIGHT && ParticleXs[i] >= ENVIRONMENT_LENGTH - 1) ||
        (dir == DOWN  && ParticleYs[i] >= ENVIRONMENT_LENGTH - 1)) {
      continue;
    }

    // If the move would cause a collision with another particle, skip this
    // particle
    if (wouldCollide(i, dir)) {
      continue;
    }

    // Commit the move
    Environment[ParticleYs[i]][ParticleXs[i]] = NONE;
    switch (dir) {
      case UP:
        ParticleYs[i]--;
        break;
      case LEFT:
        ParticleXs[i]--;
        break;
      case RIGHT:
        ParticleXs[i]++;
        break;
      case DOWN:
        ParticleYs[i]++;
        break;
    }
    Environment[ParticleYs[i]][ParticleXs[i]] = FREE;
  }
}

// Define the main function
int main() {
  // Seed the random number generator
  srandom(0);

  Init();

  for (int time = 0; time < TIME_COUNT; time++) {
    StickParticles();
    MoveRandom();
  }

  return 0;
}

